package require tclpy
package require oo::util

oo::class create ::pyman {
    variable pythonObject

    constructor {args} {
        variable pythonObject

        set className [info object class [self]]
        if {$className eq "::pyman"} {
            throw [list PYMAN LOOKUP SUBCOMMAND] {unknown subcommand must be: import or addPath}
        }

        set tclModuleName [join [lrange [split $className {:}] 0 end-2] {:}]
        set pythonModuleName [set ${tclModuleName}::pythonModuleName]
        set pythonClassName [lindex [split $className {:}] end]
        set pythonObject [string map {{::} {_}} [self]]

        ##
        ## Define a function to create the object and assign it to a global variable
        ##
        set defineInstance {def pymanDefineInstance( objName}
        set createString {globals()[objName] = }
        append createString "$pythonModuleName.${pythonClassName}("
        set argCnt [llength $args]
        for {set i 1} {$i <= $argCnt} {incr i} {
            append defineInstance ", x$i"
            if {$i > 1} {
                append createString ", "
            }
            append createString x$i
        }
        append defineInstance "):"
        append defineInstance "\n    " $createString ")"
        append defineInstance "\n    " "return"

        py eval $defineInstance
        py call pymanDefineInstance $pythonObject {*}$args
    }

    classmethod import {args} {
        switch -exact -- [llength $args] {
            1   {
                ##
                ## Just a moduleName
                ##
                set moduleName [lindex $args 0]
                set namespaceName [format {::pyman::%s} $moduleName]
            }
            3 {
                ##
                ## We have -namespace namespaceName moudleName
                ##
                if {[lindex $args 0] ne "-namespace"} {
                    throw [list PYMAN BADIMP $args] "Bad usage must be: pyman import ?-namespace namespaceName? moduleName"
                }
                set namespaceName [lindex $args 1]
                set moduleName [lindex $args 2]
            }
            default {
                ##
                ## Illegal usage
                ##
                throw [list PYMAN BADIMP $args] "Bad usage must be: pyman import ?-namespace namespaceName? moduleName"
            }
        }

        ##
        ## Format for message bodies
        ##
        set methodBodyFmt {
            variable pythonObject

            py call $pythonObject.%1$s {*}$args
        }

        ##
        ## Have python import the module
        ##
        py import $moduleName

        ##
        ## Create the Tcl namespace that maps to the imported module
        ##
        namespace eval $namespaceName {}
        set ${namespaceName}::pythonModuleName $moduleName

        ##
        ## Create the inspection macro and look at the contents of the module
        ##
        py eval "pyman.inspect = lambda : inspect.getmembers($moduleName)"
        foreach moduleItem [py call pyman.inspect] {
            lassign $moduleItem name type
            switch -glob -- $type {
                {<class *} {
                    ##
                    ## Define the class
                    ##
                    set className [format {%s::%s} [string trimright $namespaceName] $name]
                    ::oo::class create $className {superclass ::pyman}

                    ##
                    ## Remove unwanted class methods
                    ##
                    #::oo::define $className deletemethod addPath
                    #::oo::define $className deletemethod import

                    ##
                    ## Introspect to see what methods exists and add them in
                    ##
                    py eval "pyman.getMethods = lambda : inspect.getmembers($moduleName.$name)"
                    foreach classItem [py call pyman.getMethods] {
                        lassign $classItem methodName methodInfo
                        if {[string match {<unbound method *} $methodInfo] && $methodName ne "__init__"} {
                            ::oo::define $className method $methodName args [format $methodBodyFmt $methodName]
                        }
                    }
                }
                {<function *} {
                    ##
                    ## Define the wrapper procedure
                    ##
                    {*}[format {proc %1$s::%3$s {args} {py call %2$s.%3$s {*}$args}} $namespaceName $moduleName $name]
                }
            }
        }
    }

    classmethod addPath {pathName} {
        py eval "sys.path.append('$pathName')"
    }

}

py import inspect
py import sys
py import os
py import types
py eval {from types import ModuleType}
py eval {pyman = ModuleType('pyman')}
